@@ -13,15 +13,20 @@
 # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 # See the License for the specific language governing permissions and
 # limitations under the License.
+import logging
 import time
 import unittest
+import warnings
+from contextlib import contextmanager
+from copy import copy
+
+import pytest
 
 import six.moves
 
 import tenacity
-from tenacity import RetryError
-from tenacity import Retrying
-from tenacity import retry
+from tenacity import RetryError, Retrying, retry
+from tenacity.compat import make_retry_state
 
 
 class TestBase(unittest.TestCase):
@@ -33,12 +38,15 @@ class TestStopConditions(unittest.TestCase):
 
     def test_never_stop(self):
         r = Retrying()
-        self.assertFalse(r.stop(3, 6546))
+        self.assertFalse(r.stop(make_retry_state(3, 6546)))
 
     def test_stop_any(self):
-        s = tenacity.stop_any(
+        stop = tenacity.stop_any(
             tenacity.stop_after_delay(1),
             tenacity.stop_after_attempt(4))
+
+        def s(*args):
+            return stop(make_retry_state(*args))
         self.assertFalse(s(1, 0.1))
         self.assertFalse(s(2, 0.2))
         self.assertFalse(s(2, 0.8))
@@ -47,9 +55,12 @@ class TestStopConditions(unittest.TestCase):
         self.assertTrue(s(4, 1.8))
 
     def test_stop_all(self):
-        s = tenacity.stop_all(
+        stop = tenacity.stop_all(
             tenacity.stop_after_delay(1),
             tenacity.stop_after_attempt(4))
+
+        def s(*args):
+            return stop(make_retry_state(*args))
         self.assertFalse(s(1, 0.1))
         self.assertFalse(s(2, 0.2))
         self.assertFalse(s(2, 0.8))
@@ -58,7 +69,10 @@ class TestStopConditions(unittest.TestCase):
         self.assertTrue(s(4, 1.8))
 
     def test_stop_or(self):
-        s = tenacity.stop_after_delay(1) | tenacity.stop_after_attempt(4)
+        stop = tenacity.stop_after_delay(1) | tenacity.stop_after_attempt(4)
+
+        def s(*args):
+            return stop(make_retry_state(*args))
         self.assertFalse(s(1, 0.1))
         self.assertFalse(s(2, 0.2))
         self.assertFalse(s(2, 0.8))
@@ -67,7 +81,10 @@ class TestStopConditions(unittest.TestCase):
         self.assertTrue(s(4, 1.8))
 
     def test_stop_and(self):
-        s = tenacity.stop_after_delay(1) & tenacity.stop_after_attempt(4)
+        stop = tenacity.stop_after_delay(1) & tenacity.stop_after_attempt(4)
+
+        def s(*args):
+            return stop(make_retry_state(*args))
         self.assertFalse(s(1, 0.1))
         self.assertFalse(s(2, 0.2))
         self.assertFalse(s(2, 0.8))
@@ -77,24 +94,37 @@ class TestStopConditions(unittest.TestCase):
 
     def test_stop_after_attempt(self):
         r = Retrying(stop=tenacity.stop_after_attempt(3))
-        self.assertFalse(r.stop(2, 6546))
-        self.assertTrue(r.stop(3, 6546))
-        self.assertTrue(r.stop(4, 6546))
+        self.assertFalse(r.stop(make_retry_state(2, 6546)))
+        self.assertTrue(r.stop(make_retry_state(3, 6546)))
+        self.assertTrue(r.stop(make_retry_state(4, 6546)))
 
     def test_stop_after_delay(self):
         r = Retrying(stop=tenacity.stop_after_delay(1))
-        self.assertFalse(r.stop(2, 0.999))
-        self.assertTrue(r.stop(2, 1))
-        self.assertTrue(r.stop(2, 1.001))
+        self.assertFalse(r.stop(make_retry_state(2, 0.999)))
+        self.assertTrue(r.stop(make_retry_state(2, 1)))
+        self.assertTrue(r.stop(make_retry_state(2, 1.001)))
 
     def test_legacy_explicit_stop_type(self):
         Retrying(stop="stop_after_attempt")
 
-    def test_stop_func(self):
+    def test_stop_backward_compat(self):
         r = Retrying(stop=lambda attempt, delay: attempt == delay)
-        self.assertFalse(r.stop(1, 3))
-        self.assertFalse(r.stop(100, 99))
-        self.assertTrue(r.stop(101, 101))
+        with reports_deprecation_warning():
+            self.assertFalse(r.stop(make_retry_state(1, 3)))
+        with reports_deprecation_warning():
+            self.assertFalse(r.stop(make_retry_state(100, 99)))
+        with reports_deprecation_warning():
+            self.assertTrue(r.stop(make_retry_state(101, 101)))
+
+    def test_stop_func_with_retry_state(self):
+        def stop_func(retry_state):
+            rs = retry_state
+            return rs.attempt_number == rs.seconds_since_start
+
+        r = Retrying(stop=stop_func)
+        self.assertFalse(r.stop(make_retry_state(1, 3)))
+        self.assertFalse(r.stop(make_retry_state(100, 99)))
+        self.assertTrue(r.stop(make_retry_state(101, 101)))
 
 
 class TestWaitConditions(unittest.TestCase):
@@ -175,7 +205,7 @@ class TestWaitConditions(unittest.TestCase):
     def test_legacy_explicit_wait_type(self):
         Retrying(wait="exponential_sleep")
 
-    def test_wait_func_result(self):
+    def test_wait_backward_compat_with_result(self):
         captures = []
 
         def wait_capture(attempt, delay, last_result=None):
@@ -189,15 +219,18 @@ class TestWaitConditions(unittest.TestCase):
         r = Retrying(wait=wait_capture, sleep=lambda secs: None,
                      stop=tenacity.stop_after_attempt(r_attempts),
                      reraise=True)
-        self.assertRaises(Exception, r.call, dying)
+        with reports_deprecation_warning():
+            self.assertRaises(Exception, r.call, dying)
         self.assertEqual(r_attempts - 1, len(captures))
         self.assertTrue(all([r.failed for r in captures]))
 
     def test_wait_func(self):
-        r = Retrying(wait=lambda attempt, delay: attempt * delay)
-        self.assertEqual(r.wait(1, 5), 5)
-        self.assertEqual(r.wait(2, 11), 22)
-        self.assertEqual(r.wait(10, 100), 1000)
+        def wait_func(retry_state):
+            return retry_state.attempt_number * retry_state.seconds_since_start
+        r = Retrying(wait=wait_func)
+        self.assertEqual(r.wait(make_retry_state(1, 5)), 5)
+        self.assertEqual(r.wait(make_retry_state(2, 11)), 22)
+        self.assertEqual(r.wait(make_retry_state(10, 100)), 1000)
 
     def test_wait_combine(self):
         r = Retrying(wait=tenacity.wait_combine(tenacity.wait_random(0, 3),
@@ -251,7 +284,7 @@ class TestWaitConditions(unittest.TestCase):
             [tenacity.wait_fixed(8) for i in six.moves.range(1)]))
 
         for i in six.moves.range(10):
-            w = r.wait(i, 1)
+            w = r.wait(i + 1, 1)
             if i < 2:
                 self._assert_range(w, 1, 2)
             elif i < 4:
@@ -259,6 +292,30 @@ class TestWaitConditions(unittest.TestCase):
             else:
                 self._assert_range(w, 8, 9)
 
+    def test_wait_chain_multiple_invocations(self):
+        sleep_intervals = []
+        r = Retrying(
+            sleep=sleep_intervals.append,
+            wait=tenacity.wait_chain(*[
+                tenacity.wait_fixed(i + 1) for i in six.moves.range(3)
+            ]),
+            stop=tenacity.stop_after_attempt(5),
+            retry=tenacity.retry_if_result(lambda x: x == 1),
+        )
+
+        @r.wraps
+        def always_return_1():
+            return 1
+
+        self.assertRaises(tenacity.RetryError, always_return_1)
+        self.assertEqual(sleep_intervals, [1.0, 2.0, 3.0, 3.0])
+        sleep_intervals[:] = []
+
+        # Clear and restart retrying.
+        self.assertRaises(tenacity.RetryError, always_return_1)
+        self.assertEqual(sleep_intervals, [1.0, 2.0, 3.0, 3.0])
+        sleep_intervals[:] = []
+
     def test_wait_random_exponential(self):
         fn = tenacity.wait_random_exponential(0.5, 60.0)
 
@@ -305,48 +362,161 @@ class TestWaitConditions(unittest.TestCase):
         self._assert_inclusive_range(mean(attempt[8]), 28.00, 34.00)
         self._assert_inclusive_range(mean(attempt[9]), 28.00, 34.00)
 
+    def test_wait_backward_compat(self):
+        """Ensure Retrying object accepts both old and newstyle wait funcs."""
+        def wait1(previous_attempt_number, delay_since_first_attempt):
+            wait1.calls.append((
+                previous_attempt_number, delay_since_first_attempt))
+            return 0
+        wait1.calls = []
+
+        def wait2(previous_attempt_number, delay_since_first_attempt,
+                  last_result):
+            wait2.calls.append((
+                previous_attempt_number, delay_since_first_attempt,
+                last_result))
+            return 0
+        wait2.calls = []
+
+        def dying():
+            raise Exception("Broken")
+
+        retrying1 = Retrying(wait=wait1, stop=tenacity.stop_after_attempt(4))
+        with reports_deprecation_warning():
+            self.assertRaises(Exception, lambda: retrying1.call(dying))
+        self.assertEqual([t[0] for t in wait1.calls], [1, 2, 3])
+        # This assumes that 3 iterations complete within 1 second.
+        self.assertTrue(all(t[1] < 1 for t in wait1.calls))
+
+        retrying2 = Retrying(wait=wait2, stop=tenacity.stop_after_attempt(4))
+        with reports_deprecation_warning():
+            self.assertRaises(Exception, lambda: retrying2.call(dying))
+        self.assertEqual([t[0] for t in wait2.calls], [1, 2, 3])
+        # This assumes that 3 iterations complete within 1 second.
+        self.assertTrue(all(t[1] < 1 for t in wait2.calls))
+        self.assertEqual([str(t[2].exception()) for t in wait2.calls],
+                         ['Broken'] * 3)
+
+    def test_wait_class_backward_compatibility(self):
+        """Ensure builtin objects accept both old and new parameters."""
+        waitobj = tenacity.wait_fixed(5)
+        self.assertEqual(waitobj(1, 0.1), 5)
+        self.assertEqual(
+            waitobj(1, 0.1, tenacity.Future.construct(1, 1, False)), 5)
+        retry_state = make_retry_state(123, 456)
+        self.assertEqual(retry_state.attempt_number, 123)
+        self.assertEqual(retry_state.seconds_since_start, 456)
+        self.assertEqual(waitobj(retry_state=retry_state), 5)
+
+    def test_wait_retry_state_attributes(self):
+
+        class ExtractCallState(Exception):
+            pass
+
+        # retry_state is mutable, so return it as an exception to extract the
+        # exact values it has when wait is called and bypass any other logic.
+        def waitfunc(retry_state):
+            raise ExtractCallState(retry_state)
+
+        retrying = Retrying(
+            wait=waitfunc,
+            retry=(tenacity.retry_if_exception_type() |
+                   tenacity.retry_if_result(lambda result: result == 123)))
+
+        def returnval():
+            return 123
+        try:
+            retrying.call(returnval)
+        except ExtractCallState as err:
+            retry_state = err.args[0]
+        self.assertIs(retry_state.fn, returnval)
+        self.assertEqual(retry_state.args, ())
+        self.assertEqual(retry_state.kwargs, {})
+        self.assertEqual(retry_state.outcome.result(), 123)
+        self.assertEqual(retry_state.attempt_number, 1)
+        self.assertGreater(retry_state.outcome_timestamp,
+                           retry_state.start_time)
+
+        def dying():
+            raise Exception("Broken")
+        try:
+            retrying.call(dying)
+        except ExtractCallState as err:
+            retry_state = err.args[0]
+        self.assertIs(retry_state.fn, dying)
+        self.assertEqual(retry_state.args, ())
+        self.assertEqual(retry_state.kwargs, {})
+        self.assertEqual(str(retry_state.outcome.exception()), 'Broken')
+        self.assertEqual(retry_state.attempt_number, 1)
+        self.assertGreater(retry_state.outcome_timestamp,
+                           retry_state.start_time)
+
 
 class TestRetryConditions(unittest.TestCase):
 
     def test_retry_if_result(self):
-        r = (tenacity.retry_if_result(lambda x: x == 1))
+        retry = (tenacity.retry_if_result(lambda x: x == 1))
+
+        def r(fut):
+            retry_state = make_retry_state(1, 1.0, last_result=fut)
+            return retry(retry_state)
         self.assertTrue(r(tenacity.Future.construct(1, 1, False)))
         self.assertFalse(r(tenacity.Future.construct(1, 2, False)))
 
     def test_retry_if_not_result(self):
-        r = (tenacity.retry_if_not_result(lambda x: x == 1))
+        retry = (tenacity.retry_if_not_result(lambda x: x == 1))
+
+        def r(fut):
+            retry_state = make_retry_state(1, 1.0, last_result=fut)
+            return retry(retry_state)
         self.assertTrue(r(tenacity.Future.construct(1, 2, False)))
         self.assertFalse(r(tenacity.Future.construct(1, 1, False)))
 
     def test_retry_any(self):
-        r = tenacity.retry_any(
+        retry = tenacity.retry_any(
             tenacity.retry_if_result(lambda x: x == 1),
             tenacity.retry_if_result(lambda x: x == 2))
+
+        def r(fut):
+            retry_state = make_retry_state(1, 1.0, last_result=fut)
+            return retry(retry_state)
         self.assertTrue(r(tenacity.Future.construct(1, 1, False)))
         self.assertTrue(r(tenacity.Future.construct(1, 2, False)))
         self.assertFalse(r(tenacity.Future.construct(1, 3, False)))
         self.assertFalse(r(tenacity.Future.construct(1, 1, True)))
 
     def test_retry_all(self):
-        r = tenacity.retry_all(
+        retry = tenacity.retry_all(
             tenacity.retry_if_result(lambda x: x == 1),
             tenacity.retry_if_result(lambda x: isinstance(x, int)))
+
+        def r(fut):
+            retry_state = make_retry_state(1, 1.0, last_result=fut)
+            return retry(retry_state)
         self.assertTrue(r(tenacity.Future.construct(1, 1, False)))
         self.assertFalse(r(tenacity.Future.construct(1, 2, False)))
         self.assertFalse(r(tenacity.Future.construct(1, 3, False)))
         self.assertFalse(r(tenacity.Future.construct(1, 1, True)))
 
     def test_retry_and(self):
-        r = (tenacity.retry_if_result(lambda x: x == 1) &
-             tenacity.retry_if_result(lambda x: isinstance(x, int)))
+        retry = (tenacity.retry_if_result(lambda x: x == 1) &
+                 tenacity.retry_if_result(lambda x: isinstance(x, int)))
+
+        def r(fut):
+            retry_state = make_retry_state(1, 1.0, last_result=fut)
+            return retry(retry_state)
         self.assertTrue(r(tenacity.Future.construct(1, 1, False)))
         self.assertFalse(r(tenacity.Future.construct(1, 2, False)))
         self.assertFalse(r(tenacity.Future.construct(1, 3, False)))
         self.assertFalse(r(tenacity.Future.construct(1, 1, True)))
 
     def test_retry_or(self):
-        r = (tenacity.retry_if_result(lambda x: x == "foo") |
-             tenacity.retry_if_result(lambda x: isinstance(x, int)))
+        retry = (tenacity.retry_if_result(lambda x: x == "foo") |
+                 tenacity.retry_if_result(lambda x: isinstance(x, int)))
+
+        def r(fut):
+            retry_state = make_retry_state(1, 1.0, last_result=fut)
+            return retry(retry_state)
         self.assertTrue(r(tenacity.Future.construct(1, "foo", False)))
         self.assertFalse(r(tenacity.Future.construct(1, "foobar", False)))
         self.assertFalse(r(tenacity.Future.construct(1, 2.2, False)))
@@ -386,6 +556,15 @@ class TestRetryConditions(unittest.TestCase):
                           _r)
         self.assertEqual(5, r.statistics['attempt_number'])
 
+    def test_retry_if_exception_message_negative_no_inputs(self):
+        with self.assertRaises(TypeError):
+            tenacity.retry_if_exception_message()
+
+    def test_retry_if_exception_message_negative_too_many_inputs(self):
+        with self.assertRaises(TypeError):
+            tenacity.retry_if_exception_message(
+                message="negative", match="negative")
+
 
 class NoneReturnUntilAfterCount(object):
     """Holds counter state for invoking a method several times in a row."""
@@ -444,6 +623,8 @@ class NoNameErrorAfterCount(object):
 class NameErrorUntilCount(object):
     """Holds counter state for invoking a method several times in a row."""
 
+    derived_message = "Hi there, I'm a NameError"
+
     def __init__(self, count):
         self.counter = 0
         self.count = count
@@ -456,7 +637,7 @@ class NameErrorUntilCount(object):
         if self.counter < self.count:
             self.counter += 1
             return True
-        raise NameError("Hi there, I'm a NameError")
+        raise NameError(self.derived_message)
 
 
 class IOErrorUntilCount(object):
@@ -492,12 +673,14 @@ class CustomError(Exception):
         self.value = value
 
     def __str__(self):
-        return repr(self.value)
+        return self.value
 
 
 class NoCustomErrorAfterCount(object):
     """Holds counter state for invoking a method several times in a row."""
 
+    derived_message = "This is a Custom exception class"
+
     def __init__(self, count):
         self.counter = 0
         self.count = count
@@ -509,11 +692,21 @@ class NoCustomErrorAfterCount(object):
         """
         if self.counter < self.count:
             self.counter += 1
-            derived_message = "This is a Custom exception class"
-            raise CustomError(derived_message)
+            raise CustomError(self.derived_message)
         return True
 
 
+class CapturingHandler(logging.Handler):
+    """Captures log records for inspection."""
+
+    def __init__(self, *args, **kwargs):
+        super(CapturingHandler, self).__init__(*args, **kwargs)
+        self.records = []
+
+    def emit(self, record):
+        self.records.append(record)
+
+
 def current_time_ms():
     return int(round(time.time() * 1000))
 
@@ -559,6 +752,38 @@ def _retryable_test_with_unless_exception_type_no_input(thing):
     return thing.go()
 
 
+@retry(
+    stop=tenacity.stop_after_attempt(5),
+    retry=tenacity.retry_if_exception_message(
+        message=NoCustomErrorAfterCount.derived_message))
+def _retryable_test_if_exception_message_message(thing):
+    return thing.go()
+
+
+@retry(retry=tenacity.retry_if_not_exception_message(
+    message=NoCustomErrorAfterCount.derived_message))
+def _retryable_test_if_not_exception_message_message(thing):
+    return thing.go()
+
+
+@retry(retry=tenacity.retry_if_exception_message(
+    match=NoCustomErrorAfterCount.derived_message[:3] + ".*"))
+def _retryable_test_if_exception_message_match(thing):
+    return thing.go()
+
+
+@retry(retry=tenacity.retry_if_not_exception_message(
+    match=NoCustomErrorAfterCount.derived_message[:3] + ".*"))
+def _retryable_test_if_not_exception_message_match(thing):
+    return thing.go()
+
+
+@retry(retry=tenacity.retry_if_not_exception_message(
+    message=NameErrorUntilCount.derived_message))
+def _retryable_test_not_exception_message_delay(thing):
+    return thing.go()
+
+
 @retry
 def _retryable_default(thing):
     return thing.go()
@@ -675,12 +900,69 @@ class TestDecoratorWrapper(unittest.TestCase):
             self.assertTrue(isinstance(e, RetryError))
             print(e)
 
+    def test_retry_if_exception_message(self):
+        try:
+            self.assertTrue(_retryable_test_if_exception_message_message(
+                NoCustomErrorAfterCount(3)))
+        except CustomError:
+            print(_retryable_test_if_exception_message_message.retry.
+                  statistics)
+            self.fail("CustomError should've been retried from errormessage")
+
+    def test_retry_if_not_exception_message(self):
+        try:
+            self.assertTrue(_retryable_test_if_not_exception_message_message(
+                NoCustomErrorAfterCount(2)))
+        except CustomError:
+            s = _retryable_test_if_not_exception_message_message.retry.\
+                statistics
+            self.assertTrue(s['attempt_number'] == 1)
+
+    def test_retry_if_not_exception_message_delay(self):
+        try:
+            self.assertTrue(_retryable_test_not_exception_message_delay(
+                NameErrorUntilCount(3)))
+        except NameError:
+            s = _retryable_test_not_exception_message_delay.retry.statistics
+            print(s['attempt_number'])
+            self.assertTrue(s['attempt_number'] == 4)
+
+    def test_retry_if_exception_message_match(self):
+        try:
+            self.assertTrue(_retryable_test_if_exception_message_match(
+                NoCustomErrorAfterCount(3)))
+        except CustomError:
+            self.fail("CustomError should've been retried from errormessage")
+
+    def test_retry_if_not_exception_message_match(self):
+        try:
+            self.assertTrue(_retryable_test_if_not_exception_message_message(
+                NoCustomErrorAfterCount(2)))
+        except CustomError:
+            s = _retryable_test_if_not_exception_message_message.retry.\
+                statistics
+            self.assertTrue(s['attempt_number'] == 1)
+
     def test_defaults(self):
         self.assertTrue(_retryable_default(NoNameErrorAfterCount(5)))
         self.assertTrue(_retryable_default_f(NoNameErrorAfterCount(5)))
         self.assertTrue(_retryable_default(NoCustomErrorAfterCount(5)))
         self.assertTrue(_retryable_default_f(NoCustomErrorAfterCount(5)))
 
+    def test_retry_function_object(self):
+        """Test that six.wraps doesn't cause problems with callable objects.
+
+        It raises an error upon trying to wrap it in Py2, because __name__
+        attribute is missing. It's fixed in Py3 but was never backported.
+        """
+        class Hello(object):
+            def __call__(self):
+                return "Hello"
+        retrying = Retrying(wait=tenacity.wait_fixed(0.01),
+                            stop=tenacity.stop_after_attempt(3))
+        h = retrying.wraps(Hello())
+        self.assertEqual(h(), "Hello")
+
 
 class TestBeforeAfterAttempts(unittest.TestCase):
     _attempt_number = 0
@@ -688,8 +970,9 @@ class TestBeforeAfterAttempts(unittest.TestCase):
     def test_before_attempts(self):
         TestBeforeAfterAttempts._attempt_number = 0
 
-        def _before(fn, attempt_number):
-            TestBeforeAfterAttempts._attempt_number = attempt_number
+        def _before(retry_state):
+            TestBeforeAfterAttempts._attempt_number = \
+                retry_state.attempt_number
 
         @retry(wait=tenacity.wait_fixed(1),
                stop=tenacity.stop_after_attempt(1),
@@ -704,8 +987,9 @@ class TestBeforeAfterAttempts(unittest.TestCase):
     def test_after_attempts(self):
         TestBeforeAfterAttempts._attempt_number = 0
 
-        def _after(fn, attempt_number, trial_time_taken_ms):
-            TestBeforeAfterAttempts._attempt_number = attempt_number
+        def _after(retry_state):
+            TestBeforeAfterAttempts._attempt_number = \
+                retry_state.attempt_number
 
         @retry(wait=tenacity.wait_fixed(0.1),
                stop=tenacity.stop_after_attempt(3),
@@ -721,25 +1005,80 @@ class TestBeforeAfterAttempts(unittest.TestCase):
         self.assertTrue(TestBeforeAfterAttempts._attempt_number is 2)
 
     def test_before_sleep(self):
-        TestBeforeAfterAttempts._attempt_number = 0
+        def _before_sleep(retry_state):
+            self.assertGreater(retry_state.next_action.sleep, 0)
+            _before_sleep.attempt_number = retry_state.attempt_number
 
-        def _before_sleep(retry_obj, sleep, last_result):
-            self.assertGreater(sleep, 0)
-            TestBeforeAfterAttempts._attempt_number = \
-                retry_obj.statistics['attempt_number']
-
-        @retry(wait=tenacity.wait_fixed(0.1),
+        @retry(wait=tenacity.wait_fixed(0.01),
                stop=tenacity.stop_after_attempt(3),
                before_sleep=_before_sleep)
         def _test_before_sleep():
-            if TestBeforeAfterAttempts._attempt_number < 2:
+            if _before_sleep.attempt_number < 2:
                 raise Exception("testing before_sleep_attempts handler")
-            else:
-                pass
 
         _test_before_sleep()
+        self.assertEqual(_before_sleep.attempt_number, 2)
 
-        self.assertTrue(TestBeforeAfterAttempts._attempt_number is 2)
+    def test_before_sleep_backward_compat(self):
+        def _before_sleep(retry_obj, sleep, last_result):
+            self.assertGreater(sleep, 0)
+            _before_sleep.attempt_number = \
+                retry_obj.statistics['attempt_number']
+        _before_sleep.attempt_number = 0
+
+        @retry(wait=tenacity.wait_fixed(0.01),
+               stop=tenacity.stop_after_attempt(3),
+               before_sleep=_before_sleep)
+        def _test_before_sleep():
+            if _before_sleep.attempt_number < 2:
+                raise Exception("testing before_sleep_attempts handler")
+
+        with reports_deprecation_warning():
+            _test_before_sleep()
+        self.assertEqual(_before_sleep.attempt_number, 2)
+
+    def test_before_sleep_log_raises(self):
+        thing = NoIOErrorAfterCount(2)
+        logger = logging.getLogger(self.id())
+        logger.propagate = False
+        logger.setLevel(logging.INFO)
+        handler = CapturingHandler()
+        logger.addHandler(handler)
+        try:
+            _before_sleep = tenacity.before_sleep_log(logger, logging.INFO)
+            retrying = Retrying(wait=tenacity.wait_fixed(0.01),
+                                stop=tenacity.stop_after_attempt(3),
+                                before_sleep=_before_sleep)
+            retrying.call(thing.go)
+        finally:
+            logger.removeHandler(handler)
+
+        etalon_re = r'Retrying .* in 0\.01 seconds as it raised .*\.'
+        self.assertEqual(len(handler.records), 2)
+        self.assertRegexpMatches(handler.records[0].getMessage(), etalon_re)
+        self.assertRegexpMatches(handler.records[1].getMessage(), etalon_re)
+
+    def test_before_sleep_log_returns(self):
+        thing = NoneReturnUntilAfterCount(2)
+        logger = logging.getLogger(self.id())
+        logger.propagate = False
+        logger.setLevel(logging.INFO)
+        handler = CapturingHandler()
+        logger.addHandler(handler)
+        try:
+            _before_sleep = tenacity.before_sleep_log(logger, logging.INFO)
+            _retry = tenacity.retry_if_result(lambda result: result is None)
+            retrying = Retrying(wait=tenacity.wait_fixed(0.01),
+                                stop=tenacity.stop_after_attempt(3),
+                                retry=_retry, before_sleep=_before_sleep)
+            retrying.call(thing.go)
+        finally:
+            logger.removeHandler(handler)
+
+        self.assertEqual(len(handler.records), 2)
+        etalon_re = r'Retrying .* in 0\.01 seconds as it returned None'
+        self.assertRegexpMatches(handler.records[0].getMessage(), etalon_re)
+        self.assertRegexpMatches(handler.records[1].getMessage(), etalon_re)
 
 
 class TestReraiseExceptions(unittest.TestCase):
@@ -780,7 +1119,7 @@ class TestReraiseExceptions(unittest.TestCase):
 
         @retry(wait=tenacity.wait_fixed(0.1),
                stop=tenacity.stop_after_attempt(2),
-               retry=lambda x: True)
+               retry=lambda retry_state: True)
         def _mock_fn():
             calls.append('x')
 
@@ -798,7 +1137,7 @@ class TestReraiseExceptions(unittest.TestCase):
 
         @retry(wait=tenacity.wait_fixed(0.1),
                stop=tenacity.stop_after_attempt(2),
-               retry=lambda x: True,
+               retry=lambda retry_state: True,
                reraise=True)
         def _mock_fn():
             calls.append('x')
@@ -831,5 +1170,71 @@ class TestStatistics(unittest.TestCase):
         self.assertEqual(2, _foobar.retry.statistics['attempt_number'])
 
 
+class TestRetryErrorCallback(unittest.TestCase):
+
+    def setUp(self):
+        self._attempt_number = 0
+        self._callback_called = False
+
+    def _callback(self, fut):
+        self._callback_called = True
+        return fut
+
+    def test_retry_error_callback_backward_compat(self):
+        num_attempts = 3
+
+        def retry_error_callback(fut):
+            retry_error_callback.called_times += 1
+            return fut
+
+        retry_error_callback.called_times = 0
+
+        @retry(stop=tenacity.stop_after_attempt(num_attempts),
+               retry_error_callback=retry_error_callback)
+        def _foobar():
+            self._attempt_number += 1
+            raise Exception("This exception should not be raised")
+
+        with reports_deprecation_warning():
+            result = _foobar()
+
+        self.assertEqual(retry_error_callback.called_times, 1)
+        self.assertEqual(num_attempts, self._attempt_number)
+        self.assertIsInstance(result, tenacity.Future)
+
+    def test_retry_error_callback(self):
+        num_attempts = 3
+
+        def retry_error_callback(retry_state):
+            retry_error_callback.called_times += 1
+            return retry_state.outcome
+
+        retry_error_callback.called_times = 0
+
+        @retry(stop=tenacity.stop_after_attempt(num_attempts),
+               retry_error_callback=retry_error_callback)
+        def _foobar():
+            self._attempt_number += 1
+            raise Exception("This exception should not be raised")
+
+        result = _foobar()
+
+        self.assertEqual(retry_error_callback.called_times, 1)
+        self.assertEqual(num_attempts, self._attempt_number)
+        self.assertIsInstance(result, tenacity.Future)
+
+
+@contextmanager
+def reports_deprecation_warning():
+    __tracebackhide__ = True
+    oldfilters = copy(warnings.filters)
+    warnings.simplefilter('always')
+    try:
+        with pytest.warns(DeprecationWarning):
+            yield
+    finally:
+        warnings.filters = oldfilters
+
+
 if __name__ == '__main__':
     unittest.main()
